from django.urls.exceptions import Resolver404

from django_dynamic_path import DynamicPath
from django_early_return import EarlyReturn

from .path_resolver import PathResolver
from .view import view, preprocess

def make_tree_view(root_module_name="view_tree"):
    '''
        Return an object suitable for use in django's urlpatterns.

        root_module_name should be the dotted python path to a python package containing your "view tree"
    '''
    return DynamicPath(
        PathResolver(root_module_name),
        view,
    )



'''
module_tree.py

Implements ModuleTree -> a class which walks a python package, creating an easy-to-traverse tree structure.
'''
from importlib import import_module
from os import path
from pathlib import Path

class ConfigurationError(Exception):
    pass

class ViewTreeNode:
    '''
        Represents a node (directory) in a view tree.

        self.view_tree_path
            the path of this directory, relative to the root of the view tree
        self.module
            the view_tree_node.py module
            will be None if this module does not have a view_tree_node.py file
        self.subtrees
            a dictionary of ViewTreeNodes representing subtrees of this one
    '''
    def __init__(self, view_tree_path, package_name, package_path):
        self.view_tree_path = view_tree_path

        try :
            self.module = import_module(package_name +'.view_tree_node')
        except ModuleNotFoundError :
            self.module = None

        self.subtrees = {
            subdir.name: ViewTreeNode(
                path.join(view_tree_path, subdir), 
                f'{package_name}.{subdir.name}', 
                subdir,
            )
            # Note - pretty well any directory name can be imported dynamically with import_module, but directories containing '.' cannot, because a.b is treated as a/b by the import system
            # Also exclude any directory beginning with '__' -> mainly to exclude __pycache__ directory
            for subdir in package_path.iterdir() 
            if subdir.is_dir() and not '.' in subdir.name and not subdir.name.startswith('__')
        }

def _get_valid_package_directory(module):
    try :
        f = module.__file__
    except AttributeError :
        return
    if f is None :
        return
    p = Path(f)
    if p.name != '__init__.py' :
        return
    return p.parent

class ViewTree(ViewTreeNode):
    def __init__(self, package_name):
        # Note - if import error occurs, that should be pretty self-explanatory to the user - no custom message needed
        module = import_module(package_name)
        p = _get_valid_package_directory(module)
        if not p :
            raise ConfigurationError('The root of a module tree must be a python package containing an __init__.py file')
        return super().__init__('', package_name, p)


import functools
import os

from django import http
from django_early_return import EarlyReturn
from django.views.decorators.csrf import csrf_exempt

import django_referer_csrf

http_method_names = ['get', 'post', 'put', 'patch', 'delete', 'head', 'options', 'trace']

def preprocess(request, handlers):
    branch_data = {}
    for node, arg in handlers :
        try :
            p = getattr(node.module, 'preprocess')
        except AttributeError :
            # Modules don't have to implement preprocess()
            continue

        # Not all modules capture an arg from the url
        args = [arg] if arg is not None else []

        r = p(request, *args, **branch_data)
        # preprocess can return a dictionary to be merged into the "branch data"
        if r :
            branch_data.update(r)

    return branch_data

'''
    Note - we mark the view as csrf_exempt, and then provide our own CSRF protection.
    This is so that users can turn CSRF protection off for particular endpoints.
'''
@csrf_exempt
def view(request, *handlers):
    method = request.method.lower()
    handler_node = handlers[-1][0]
    handler_module = handler_node.module

    # Apply CSRF protection, unless the handler module has set CSRF_EXEMPT=True 
    if not getattr(handler_module, 'CSRF_EXEMPT', False) :
        if not django_referer_csrf.is_valid(request) :
            return http.HttpResponseForbidden('CSRF check failed.')

    if method not in http_method_names :
        return _method_not_allowed(request, handler_module)

    try :
        handler_func = _get_handler_func(handler_module, method)
    except AttributeError :
        if method == 'options' :
            # provide default options implementation
            return _options(request, handler_module)
        return _method_not_allowed(request, handler_module)

    # We recommend storing templates in same directory as the handler module.
    # You'll then want to put the root directory of your tree view in your template DIRS
    # This property enables you to prepend the appropriate path to a template name (relative to the root of your tree view).
    # This will either be empty string (for the root of the view tree) or end in '/'
    request.view_tree_dir = os.path.join(handler_node.view_tree_path, '')

    # Include this, directly, for backward compatibility
    request.view_tree_path = handler_node.view_tree_path

    # Run all preprocess functions
    '''
        Note - we expect you to perform any required authentication in your preprocess functions.
        If authentication fails, we recommend raising a django_early_return.EarlyReturn exception.
        That way you can use our is_get_allowed(path, user) helper method to determine if a given user is currently allowed to access a given path.
    '''
    branch_data = preprocess(request, handlers)

    # Run the view
    return handler_func(request, **branch_data)

    # TODO:
    # Should we allow modules to implement a postprocess(response, request), which runs in reverse order? This could be used, for example, to set response headers for an entire branch

'''
    Add a special "test_page_visibility" property to the view
    This is to support django_page_visibility (which is not yet published)

    This function should raise an exception if the user is not currently allowed to view the page.
    Note that our implemenation of this requires users to all of their permission checks inside preprocess() functions.
'''
def test_page_visibility(request, *handlers):
    preprocess(request, handlers)
view.test_page_visibility = test_page_visibility

def _get_handler_func(handler_module, method):
    '''Get final handler function, or raise raise AttributeError'''
    try :
        return getattr(handler_module, method)
    except AttributeError :
        pass

    if method == 'head' :
        return getattr(handler_module, 'get')

    raise AttributeError()

def _method_not_allowed(request, handler_module):
    return http.HttpResponseNotAllowed(_allowed_methods(handler_module))
def _options(request, handler_module):
    response = http.HttpResponse()
    response['Allow'] = ', '.join(_allowed_methods(handler_module))
    response['Content-Length'] = '0'
    return response
def _allowed_methods(handler_module):
    methods = [m for m in http_method_names if hasattr(handler_module, m)]
    if 'options' not in methods :
        methods.append('options')
    return methods


from django.conf import settings

from .view_tree import ViewTree

class PathResolver:
    class NoMatch(Exception):
        pass

    def __init__(self, root_module_name):
        self.root_module_name = root_module_name
        self.view_tree = ViewTree(root_module_name)

    def __call__(self, path):
        try :
            return (self.get_handler_list(path), {})
        except self.NoMatch :
            pass

        if not settings.DEBUG :
            return

        try :
            '''
                Reload the entire module tree, in case a new module has been added
                (django's runserver won't auto-reload when adding new modules, since you don't have to update existing files to import them)

                Note - this does not catch:
                    - modules that were deleted (cached, same response is returned)
                    - directory already existed, you received 404, then you added __init__.py
            '''
            self.view_tree = ViewTree(self.root_module_name)
            return (self.get_handler_list(path), {})
        except self.NoMatch :
            pass

    def get_handler_list(self, path):
        '''
            Returns a list of (view_tree_node, captured_argument) tuples.

            Will raise NoMatch if path cannot be mapped to a final view tree node (with module != None).

            Note: path should not have a leading '/'.
            If it does, we consider the path to start with an empty segment (which can still map to a string__ handler module).
        '''
        previous_node = self.view_tree
        handler_list = [(previous_node, None)]

        while path :
            '''
                Notice - the path must end in 
            '''

            '''
                Note: the path must end with / for the last segment to match a "segment handler" (ie. fixed, int__, or str__).
                If the path does not end with a /, the last segment can only match a path__ handler.

                This is by design. We don't want multiple urls (ie. with and without trailing /) to map to the same handler. Trailing / are basically necessary in urls for relative links to work.

                TODO - if url is missing trailing '/', but would have matched otherwise, should we redirect? Provide this as an option? That's how Django works.
            '''
            try :
                segment, rest = path.split('/', 1)
            except ValueError :
                pass
            else:
                # Does the segment match a submodule name exactly?
                try :
                    node = previous_node.subtrees[segment]
                except KeyError :
                    pass
                else :
                    handler_list.append((node, None))
                    path = rest
                    previous_node = node
                    continue

                # Are there any "capturing" subtrees defined?
                try :
                    node = previous_node.subtrees['int__']
                except KeyError :
                    pass
                else :
                    try :
                        arg = int(segment)
                    except ValueError :
                        pass
                    else :
                        handler_list.append((node, arg))
                        path = rest
                        previous_node = node
                        continue

                try :
                    node = previous_node.subtrees['string__']
                except KeyError :
                    pass
                else :
                    handler_list.append((node, segment))
                    path = rest
                    previous_node = node
                    continue

            # If they have a path__ module defined, consume the entire remaining path
            try :
                node = previous_node.subtrees['path__']
            except KeyError :
                pass
            else :
                handler_list.append((node, path))
                path = ''
                previous_node = node
                continue

            raise self.NoMatch()

        if not handler_list[-1][0]:
            raise self.NoMatch()

        return handler_list
